import time
import random
from typing import Tuple, Iterator, Iterable

import torch
from torch import Tensor
from datasets import Dataset

from optimus.tokenizers import SentencePieceTokenizer


class _OptimusDLIter(Iterator):
    def __init__(self, dl):
        """
        _OptimusDL iterator.

        """
        self.dl = dl
        self.curr = 0

    def __next__(self) -> Tuple[Tensor, Tensor]:
        if self.curr > len(self.dl) - 1:
            raise StopIteration

        x = self.dl._items[:, self.curr * self.dl.seq_len     : (self.curr + 1) * self.dl.seq_len]
        y = self.dl._items[:, self.curr * self.dl.seq_len + 1 : (self.curr + 1) * self.dl.seq_len + 1]

        self.curr += 1

        return x, y

class _OptimusDL(Iterable):
    def __init__(self, ds, tok, bs, seq_len, shuffle, device):
        """
        See 'OptimusDataLoader'.

        """
        self.ds = ds
        self.tok = tok
        self.bs = bs
        self.seq_len = seq_len
        self.shuffle = shuffle
        self.device = device

        start = time.time()
        print("Tokenizing dataset...")

        # tokenize the dataset, add BOS/EOS tokens
        self._data = [torch.tensor(self.tok.encode(item, bos=True, eos=True),
                                   dtype=torch.long) for item in self.ds]

        print(f"Done. Took {time.time() - start:.2f}s.")

        # pre-calculate the number of batches in the dataset

        # Note: there's a special case we need to be careful about; since the
        # predictions are simply the inputs shifted to the right by one value;
        # there's a case when the dataset ends before we can get these
        # shifted-right predictions; this occurs iff `batch_len % seq_len == 0`;
        # to avoid this, we have to be explicit about the available number of
        # batches (by simply subtracting 1 from the total number of available
        # batches)
        dataset_stream_len = 0
        for sample in self._data:
            dataset_stream_len += len(sample)

        batch_len = dataset_stream_len // self.bs
        self.num_batches = batch_len // self.seq_len

        if batch_len % self.seq_len == 0:
            self.num_batches -= 1

    def _process_data_before_iter(self):
        data = self._data

        # shuffle text (this keeps information intact, as documents are
        # shuffled, not the tokens inside of them!)
        if self.shuffle:
            random.shuffle(data)

        # stack the whole dataset into one big line of text
        data = torch.cat(data, dim=-1)

        # make streams (drop elements which don't cleanly fit)
        batch_len = data.shape[0] // self.bs
        data = data[:batch_len * self.bs]
        data = data.view(self.bs, batch_len).contiguous()

        # move data to device
        self._items = data.to(self.device)

    def to(self, device) -> None:
        """
        See 'OptimusDataLoader.to(device)'.

        """
        self.device = device

    def __iter__(self) -> Iterator[_OptimusDLIter]:
        """
        Return an iterator over the dataloader object.

        """
        self._process_data_before_iter()
        return _OptimusDLIter(self)

    def __len__(self) -> int:
        return self.num_batches


class OptimusDataLoader():

    def __init__(self,
                 train_ds: Dataset,
                 test_ds: Dataset,
                 tok: SentencePieceTokenizer,
                 bs: int,
                 seq_len: int,
                 shuffle: bool = True,
                 device: str | torch.device = 'cpu'):
        """
        A data loader specialized for training the Optimus model. This is mostly
        whatever pytorch is doing with the default DataLoader, but has a few
        extra bits specific to NLP tasks. Essentially, it is a stripped down
        version of fastai's DataLoaders object.

        The dataloader should receive text unformatted and untokenized. This is
        all taken care of inside the dataloader. Shuffling, tokenization and
        batching all happens when calling iter() on the dataloader's train and
        test attributes. This should happen before each epoch while training.

        Args:
            train_ds (Dataset): The training dataset.
            test_ds (Dataset): The testing dataset.
            tok (SentencePieceTokenizer): Tokenizer to use.
            bs (int): Batch size. The number of samples per batch to load.
            seq_len (int): Sequence length. Also referred to as context length.
            shuffle (bool): Whether to shuffle the training data before
                training. True means shuffle before every epoch. Defaults to
                True.
            device (str | torch.device): The device where to put the data (CPU,
                GPU or other devices). Defaults to CPU.

        """
        self.train = _OptimusDL(ds=train_ds,
                                tok=tok,
                                bs=bs,
                                seq_len=seq_len,
                                shuffle=shuffle,
                                device=device)

        self.test = _OptimusDL(ds=test_ds,
                               tok=tok,
                               bs=bs,
                               seq_len=seq_len,
                               shuffle=False,
                               device=device)

    def to(self, device: str | torch.device) -> 'OptimusDataLoader':
        """
        Move data to device. This creates a copy on the specified device (if the
        device is different from the one the data currently resides on).

        Args:
            device(str | torch.device): Device to move data to.

        Returns:
            OptimusDataLoader: The dataloader moved to the new device.

        """
        self.train.to(device)
        self.test.to(device)
        return self

    def cpu(self) -> 'OptimusDataLoader':
        return self.to(device=torch.device('cpu'))

    def cuda(self) -> 'OptimusDataLoader':
        return self.to(device=torch.device('cuda'))

    def __iter__(self):
        raise TypeError("'OptimusDataLoader' is not iterable. Please use "
                        "iter(OptimusDataLoader.train) or "
                        "iter(OptimusDataLoader.test) to get an iterable for "
                        "the train or test dataloaders, respectively")
