import os
from typing import List
from collections.abc import Iterable

import sentencepiece as spm


class SentencePieceTokenizer():
    def __init__(self, model_path: str = 'optimus.model'):
        """
        SentencePiece tokenizer.

        Args:
            model_path (str): Path of the tokenizer model. Defaults to
                'optimus.model'.

        """

        assert os.path.exists(model_path), \
            f"{model_path} is not a valid tokenizer model file!"

        self.sp_model = spm.SentencePieceProcessor(model_file=model_path)

        self.unk_id = self.sp_model.unk_id() # 0
        self.bos_id = self.sp_model.bos_id() # 1
        self.eos_id = self.sp_model.eos_id() # 2
        self.pad_id = self.sp_model.pad_id() # -1 (no pad token)

        assert self.sp_model.vocab_size() == self.sp_model.get_piece_size()

    @classmethod
    def train(cls, text_iter: Iterable[List[str]], vocab_size: int = 32000) -> None:
        """
        Train a SentencePiece tokenizer on a corpus of text. The text should be
        passed in as lines through 'text_iter'. For example, if the text is
        contained in a file called 'corpus.txt', the python code to train the
        model is:

            ```python
            file = 'corpus.txt'
            with open(file, 'r') as f:
                lines = f.readlines()
                Tokenizer.train(iter(lines), vocab_size=32000)
            ```

        Additionally, multiple files can be opened, and their content
        concatenated before being passed as an iterable.

        Args:
            text_iter (Iterable[List[str]]): Iterator over a list of strings,
                which is the lines contained in the file(s). This text is used
                to built the tokenizer.
            vocab_size (int): The size of the final vocabulary.

        """
        spm.SentencePieceTrainer.train(sentence_iterator=text_iter,
                                       model_prefix='optimus',
                                       vocab_size=vocab_size,
                                       max_sentence_length=4096,
                                       input_sentence_size=1000000,
                                       shuffle_input_sentence=True,
                                       remove_extra_whitespaces=False,
                                       normalization_rule_name='identity',
                                       model_type='unigram')

    def encode(self, input: str, bos: bool, eos: bool) -> List[int]:
        """
        Encode the string passed as argument to numerical id's, according to the
        vocabulary.

        Args:
            input (str): The string to encode.
            bos (bool): Whether to prepend the BOS (beginning of stream) token.
            eos (bool): Whether to append the EOS (end of stream) token.

        Returns: The list of encoded id's.

        """
        assert type(input) is str
        ids: List[int] = self.sp_model.encode_as_ids(input)
        if bos:
            ids = [self.bos_id] + ids
        if eos:
            ids = ids + [self.eos_id]
        return ids

    def encode_as_pieces(self, input: str) -> List[str]:
        """
        Encode the string passed as argument to vocab tokens, according to the
        vocabulary.

        Args:
            input (str): The string to encode.

        Returns: A list of tokens.

        """
        return self.sp_model.encode_as_pieces(input)

    def decode(self, input: List[int]) -> str:
        """
        Decode the list of id's passed as argument back to a string, according
        to the vocabulary.

        Args:
            input (List[int]): The list of id's to decode.

        Returns: A string, decoded from the id's.

        """
        return self.sp_model.decode(input)

    def __len__(self) -> int:
        """
        Return the vocabulary size.
        """
        return self.sp_model.vocab_size()


if __name__=='__main__':

    train = False

    if train is True:
        filename = './wikitext-103/wiki.train.tokens'
        with open(filename, 'r') as f:
            lines = f.readlines()
            SentencePieceTokenizer.train(iter(lines), vocab_size=16000)

    else:
        tok = SentencePieceTokenizer(model_path='./optimus.model')
        print(len(tok))
        print(tok.encode("this is some sunny day", False, True))
        print(tok.encode_as_pieces("this is some sunny day"))
        print(tok.decode([1, 77, 34, 122, 9, 5, 10181, 206, 2]))
